from typing import Optional

from tensorflow.python.keras import Model
from tensorflow.python.keras.applications import ResNet50
from tensorflow.python.keras.layers import Flatten, Dense
from tensorflow.python.keras.losses import categorical_crossentropy
from tensorflow.python.keras.metrics import categorical_accuracy
from tensorflow.python.keras.optimizers import sgd, nadam, adam
from tensorflow.python.keras.models import load_model


def get_new_model(n_classes, height, width, channels=3, dense=512, lr=1e-5, weights='imagenet', pooling='avg'):
    """
    Creates a new ResNet model.

    :param n_classes: the number of classes.
    :param height: the image height.
    :param width: the image width.
    :param channels: the number of channels in the image.
    :param dense: the number of units in the dense layer before the output layer.
    :param lr: the learning rate of the model. If set to None the model is not compiled.
    :param weights: the initial weights to use. Either 'imagenet', None or the path to the weights file to be loaded.
    :param pooling: the pooling to use on the last layer. One of None, 'avg', or 'max'
    :return: the new model.
    """
    rn50 = ResNet50(include_top=False, weights=weights, input_shape=(height, width, channels), pooling=pooling)
    top = Flatten()(rn50.output)

    if dense is not None:
        top = Dense(dense)(top)

    top = Dense(n_classes, activation='softmax')(top)
    m = Model(inputs=rn50.input, outputs=top)
    if lr is not None:
        m.compile(adam(lr=lr), categorical_crossentropy, [categorical_accuracy])
    return m


def get_saved_model(path, model: Optional[Model] = None):
    """
    Restores a model from save.

    :param path: the path to the saved model.
    :param model: the model to restore weights into.
                  If None the method assumes that the model was saved entirely (not just weights).
                  If not None the method tries to load saved weights into the supplied model.
    :return: the restored model.
    """
    if model is None:  # assumes entire model is saved
        return load_model(path)
    else:  # assumes only weights saved
        model.load_weights(path)
        return model
