from typing import List, Union

from segmentation.gensim import GensimModel
from segmentation.imagesize import ImageSizeModel
from segmentation.classifier import ClassifierModel
from segmentation.ngram import NGramModel
from segmentation.pagenumber import PageNumberModel
from segmentation.segmentationmodel import SegmentationModel
from util.page import Page


class MultiModel(SegmentationModel):
    """
    A suggestion for a combined model which has had success in our testing.
    """

    def __init__(self):
        self.page_number = PageNumberModel()
        self.image_size = ImageSizeModel()
        self.ngram1w = NGramModel(word=True, n=1)
        self.ngram2w = NGramModel(word=True, n=2)
        self.ngram3w = NGramModel(word=True, n=3)
        self.ngram1c = NGramModel(word=False, n=1)
        self.ngram2c = NGramModel(word=False, n=2)
        self.ngram3c = NGramModel(word=False, n=3)
        self.gensim = GensimModel()
        self.multipred = ClassifierModel()

        # Default values are probably not optimal, it is recommended to use the train method before using the model
        self.sieve = [
            (self.page_number, 0.5),
            (self.ngram3c, 0.25),
            (self.ngram2c, 0.40),
            (self.ngram1c, 0.65),
            (self.ngram3w, 0.05),
            (self.ngram2w, 0.08),
            (self.ngram1w, 0.15),
            (self.gensim, 0.25),
            (self.image_size, 0.5),
            (self.multipred, 0.94)
        ]

        self.correct = [0] * len(self.sieve)
        self.wrong = [0] * len(self.sieve)

    # Overrides
    def predict(self, page1: Page, page2: Page):
        res = 0
        i = 0
        for model, threshold in self.sieve:
            res = model.predict(page1, page2)
            if res > threshold:
                if page1.belongs_with(page2):
                    self.correct[i] += 1
                else:
                    self.wrong[i] += 1
                return 1.0
            i += 1

        return res

    def train(self, pages: List[Union[Page, List[Page]]]):
        """
        Finds the best threshold for each model.

        :param pages: the pages to train with.
        """
        new_sieve = []
        for i in range(len(self.sieve)):
            m = self.sieve[i][0]
            t, c = m.best_threshold(pages, 'ajr')

            new_sieve.append(((m, t), c))

        def key(x):
            return x[1].ppv()

        new_sieve.sort(key=key)
        self.sieve = [ns[0] for ns in new_sieve]
