import re
from collections import Counter

from segmentation.segmentationmodel import SegmentationModel
from util.page import Page
from util.preprocess import preprocess


class NGramModel(SegmentationModel):
    """
    Uses n-grams to compute similarity between pages.
    """

    def __init__(self, word: bool = True, n: int = 1, preprocess: bool = False, sep=r"\s+"):
        """
        :param word: whether or not to use word n-grams or character n-grams
        :param n: the number of characters/words to use.
        :param preprocess: whether or not to preprocess strings before comparing.
        :param sep: the regex separator to use. Only applies to words.
        """
        self.word = word
        self.n = n
        self.preprocess = preprocess
        self.regex = re.compile(sep)

    # Overrides
    def predict(self, page1: Page, page2: Page) -> float:
        s1 = page1.text.lower()
        s2 = page2.text.lower()

        if self.preprocess:
            s1 = preprocess(s1)
            s2 = preprocess(s2)

        if self.word:
            res = self.compare_word(s1, s2)
        else:
            res = self.compare(s1, s2)

        return res

    def create_set(self, s):
        st = Counter()
        for i in range(len(s) - self.n + 1):
            sub = s[i:i + self.n]
            if isinstance(sub, list):
                sub = tuple(sub)
            st[sub] += 1
        return st

    def compare(self, s1, s2):
        set1 = self.create_set(s1)
        if len(set1) == 0:
            return 0

        set2 = self.create_set(s2)
        if len(set2) == 0:
            return 0

        intersection = set1 & set2

        common = sum(intersection.values())
        return common / (sum(set1.values()) + sum(set2.values()) - common)

    def compare_word(self, s1, s2):
        return self.compare(self.regex.split(s1), self.regex.split(s2))
