import re
from typing import Optional, List, Union

from segmentation.segmentationmodel import SegmentationModel
from util.page import Page


class PageNumberModel(SegmentationModel):
    """
    A class that predicts pages using a regular expression to find page numbers.
    """

    def __init__(self, use_entire_range: bool = False, pattern: Optional[str] = None):
        if pattern is None:
            pattern = r"\d+[^\n\w]*av[^\n\w]*\d+"
            if use_entire_range:
                pattern = r"side ?" + pattern
        self.regex = re.compile(pattern, re.IGNORECASE)
        self.regex2 = re.compile(r"\d+", re.IGNORECASE)
        if use_entire_range:
            RuntimeWarning("Using entire range is experimental and not fully implemented")
        self.use_entire_range = use_entire_range

    def get_page_nr(self, page):  # finds the page numbers
        txt = page.text
        hits = self.regex.findall(txt)
        if len(hits) == 0:
            return None, None
        else:
            h = self.regex2.findall(hits[0])
            page = int(h[0])
            of = int(h[1])
            return page, of

    # Overrides
    def predict(self, page1: Page, page2: Page) -> float:
        p1, o1 = self.get_page_nr(page1)
        p2, o2 = self.get_page_nr(page2)

        if self.use_entire_range:
            if p1 is None:
                if p2 is None:
                    return 0
                else:
                    if 1 < p2 <= o2:
                        return 1
                    else:
                        return 0
            else:
                if p2 is None:
                    if 1 <= p1 < o1:
                        return 1
                    else:
                        return 0
                else:
                    if 1 < o1 == o2 and p1 < p2:
                        return 1
                    return 0
        else:
            if p1 is None or p2 is None:
                return 0

            if 1 < o1 == o2 and p1 < p2:
                return 1

            return 0

    # Overrides
    def predict_batch(self, pages: List[Union[Page, List[Page]]]) -> (list, list):
        # Only differs from super method if use_entire_range is True.

        if isinstance(pages[0], list):
            pred = []
            ans = []
            for mother in pages:
                res = self.predict_batch(mother)
                pred += res[0]
                ans += res[1]
            return pred, ans
        else:
            if self.use_entire_range:
                pred = []
                i = 0
                while i < len(pages) - 1:
                    p, o = self.get_page_nr(pages[i])
                    if p is not None and 1 <= p <= o > 1:
                        for k in range(i - 1, i - p):
                            if k >= 0:
                                pred[k] = 1
                        for k in range(i, i + o - p + 1):
                            if k < len(pages):
                                pred.append(1)
                        i += o - p
                    else:
                        pred.append(0)
                    i += 1
                ans = [pages[i].belongs_with(pages[i + 1]) for i in range(len(pages) - 1)]
                return pred, ans
            else:
                return super().predict_batch(pages)
