from math import inf
from typing import List, Union, Tuple

import numpy as np

from util.confusion import BinaryConfusionMatrix
from util.page import Page


class SegmentationModel:
    def predict(self, page1: Page, page2: Page) -> float:
        """
        Predicts using the model.
        The output returns as a float rather than a boolean so that confidence of the output is not lost.

        :param page1: the first page to use for the prediction.
        :param page2: the second page to use for the prediction
        :return: the prediction in float form, where 0 is not belonging and 1 is belonging.
        """
        raise NotImplementedError

    def predict_batch(self, pages: List[Union[Page, List[Page]]]) -> (list, list):
        """
        Predicts a group of pages at once. By default it just compares two and two pages.

        :param pages: a list of pages or list of lists of pages to predict the segmentation for.
        :return: a list of predictions and answers of equal length ordered based on the input.
        """

        if isinstance(pages[0], Page):
            pred = [0.0] * (len(pages) - 1)
            ans = [0.0] * len(pred)
            for i in range(len(pred)):
                pred[i] = self.predict(pages[i], pages[i + 1])
                ans[i] = pages[i].belongs_with(pages[i + 1])
            return pred, ans
        else:  # list
            pred = []
            ans = []
            i = 0
            for mother in pages:
                i += 1
                p, a = self.predict_batch(mother)
                pred += p
                ans += a
            return pred, ans

    def get_confusion_matrix(self, pages: List[Page], threshold: float = 0.5) -> BinaryConfusionMatrix:
        """
        Creates a confusion matrix using this model.

        :param pages: the pages to predict.
        :param threshold: the threshold to use.
        :return: the confusion matrix resulting from the predictions.
        """
        pre, ans = self.predict_batch(pages)
        conf = BinaryConfusionMatrix()
        conf.add(pre, ans, threshold=threshold)

        return conf

    def get_multiple_cm(self, pages: List[Page], start: float = 0, end: float = 1, n_steps: int = 100) -> \
            List[Tuple[float, BinaryConfusionMatrix]]:
        """
        Gets multiple confusion matrices for varying threshold values.

        :param pages: the pages to test.
        :param ans: the correct answers for the inputs.
        :param start: the starting threshold.
        :param end: the end threshold.
        :param n_steps: the number of steps between the starting and ending threshold.
        :return: a list containing tuples with the threshold and the corresponding confusion matrix.
        """
        confs = []
        preds, ans = self.predict_batch(pages)
        for t in np.arange(start, end, (end - start) / n_steps):
            c = BinaryConfusionMatrix()
            c.add(preds, ans, t)
            confs.append((t, c))
        return confs

    def best_threshold(self, pages: List[Page], method: str) -> (float, float):
        """
        Finds the best threshold for a specific list of pages based on the specified method.

        :param pages: the list of pages to use for the computation.
        :param method: the BinaryConfusionMatrix method to use.
        :return: the best threshold along with its corresponding confusion matrix.
        """
        confs = self.get_multiple_cm(pages)
        mx = -inf
        best_t = -inf
        best_c = None
        for t, c in confs:
            res = getattr(c, method)()
            if res > mx:
                mx = res
                best_t = t
                best_c = c
        return best_t, best_c

    @staticmethod
    def get_ans(pages: List[Page]) -> list:
        return [pages[i].belongs_with(pages[i + 1]) for i in range(len(pages) - 1)]
