import math
import re
from collections import Iterable
from itertools import zip_longest
from math import log
from typing import Union


class BinaryConfusionMatrix:
    """
    Class representing a confusion matrix for two classes.
    The add method is used to automatically increment the different predictive values.
    """

    def __init__(self, tp=0, fp=0, fn=0, tn=0):
        self.tp = tp
        self.fp = fp
        self.fn = fn
        self.tn = tn

    def add(self, pre: Union[Iterable, bool, float, int], ans: Union[Iterable, bool, float, int],
            threshold: float = 0.5) -> None:
        """
        Adds a new prediction and answer into the confusion matrix.

        :param pre: the predicted value(s).
        :param ans: the actual value(s).
        :param threshold: the threshold value, a prediction value above this threshold counts as positive prediction.
        """
        if pre is None or ans is None:  # not same length, or invalid input
            raise IndexError
        if isinstance(pre, Iterable):
            if isinstance(ans, Iterable):
                for a, b in zip_longest(pre, ans):
                    self.add(a, b, threshold)
            else:
                raise ValueError
        else:  # should cover bool, float and int
            if pre > threshold:
                if ans > threshold:
                    self.tp += 1
                else:
                    self.fp += 1
            else:
                if ans > threshold:
                    self.fn += 1
                else:
                    self.tn += 1

    def sum(self):
        return self.tp + self.fp + self.fn + self.tn

    def acc(self) -> float:  # accuracy
        return div((self.tp + self.tn), self.sum())

    def positive(self) -> float:  # number of positive values
        return self.tp + self.fn

    def negative(self) -> float:  # number of negative values
        return self.tn + self.fp

    def prevalence(self) -> float:  # rate of positive values
        return div(self.positive(), self.sum())

    def predicted_pos(self) -> float:  # number of predicted positive values
        return self.tp + self.fp

    def predicted_neg(self) -> float:  # number of predicted negative values
        return self.tn + self.fn

    def tpr(self) -> float:  # true positive rate
        return div(self.tp, self.positive())

    def tnr(self) -> float:  # true negative rate
        return div(self.tn, self.negative())

    def fpr(self) -> float:  # false positive rate
        return div(self.fp, self.negative())

    def fnr(self) -> float:  # false negative rate
        return div(self.fn, self.positive())

    def ppv(self) -> float:  # positive predictive value
        return div(self.tp, self.predicted_pos())

    def fdr(self) -> float:  # false discovery rate
        return div(self.fp, self.predicted_pos())

    def for_(self) -> float:  # false omission rate
        return div(self.fn, self.predicted_neg())

    def npv(self) -> float:  # negative predictive value
        return div(self.tn, self.predicted_neg())

    def plr(self) -> float:  # positive likelihood ratio
        return div(self.tpr(), self.fpr())

    def nlr(self) -> float:  # negative likelihood ratio
        return div(self.fnr(), self.tnr())

    def dor(self) -> float:  # diagnostic odds ratio
        return div(self.tp * self.tn, self.fp * self.fn)

    def f_score(self, beta=0.1) -> float:  # f-score
        return div((1 + beta * beta) * self.tp, ((1 + beta * beta) * self.tp + beta * beta * self.fn + self.fp))

    def mcc(self) -> float:  # Matthews correlation coefficient
        return div((self.tp * self.tn - self.fp * self.fn),
                   math.sqrt((self.tp + self.fp) * (self.tp + self.fn) * (self.tn + self.fp) * (self.tn + self.fn)))

    def bm(self) -> float:  # informedness
        return self.tpr() + self.tnr() - 1

    def mk(self) -> float:  # markedness
        return self.ppv() + self.npv() - 1

    def ajr(self, base=0.95, value=0.5):
        # made such that if ppv = base, then the result is mcc * value
        return self.mcc() * self.ppv() ** (log(value) / log(base))

    def all(self):
        attr = [s for s in dir(self) if not re.match(".*__.*|add|all", s)]  # all public attributes except add and all
        return {a: getattr(self, a)() if callable(getattr(self, a)) else getattr(self, a) for a in attr}

    def __str__(self) -> str:
        return "[TP: %s, FP: %s, FN: %s, TN: %s]" % (self.tp, self.fp, self.fn, self.tn)


def div(a, b):
    # Simple function to prevent division by zero
    if b == 0:
        return 0
    return a / b
