import json
import os
from numbers import Number
from typing import Union, List, Optional, Tuple

import pandas as pd
import tensorflow as tf
from tensorflow.python.keras.backend import one_hot

from util.image import process_image
from util.page import Page

PARALLEL_CALLS = 16
BUFFER_SIZE = 8


def read_csv(filepath, delimiter, ncolumns=None, headers=True):
    """
    Read csv file and create a dataframe, with the added option of selecting number of columns.

    :param filepath: Path to csv file.
    :param delimiter: The character to split with.
    :param ncolumns: Number of columns in csv. Every delimiter after the last one is interpreted as a single string.
    :param headers: A boolean value representing whether the .csv-file has headers.
    :return: A dataframe containing the data from the .csv-file.
    """

    file = open(filepath, "r", encoding="utf8")
    lines = []

    if headers is True:
        if ncolumns is not None:
            columns = file.readline().split(delimiter, ncolumns - 1)
            columns[ncolumns - 1] = columns[ncolumns - 1].replace("\n", "")
        else:
            columns = file.readline().split(delimiter)
            ncolumns = len(columns)
            columns[ncolumns - 1] = columns[ncolumns - 1].replace("\n", "")
    else:
        firstline = file.readline().split(delimiter)
        ncolumns = len(firstline)
        lines.append(firstline)

    line = file.readline().replace("\n", "").split(delimiter, ncolumns - 1)
    i = 0
    while len(line) > 1:
        i += 1
        lines.append(line)
        line = file.readline().replace("\n", "").split(delimiter, ncolumns - 1)

    df = pd.DataFrame(lines, index=None)
    df.append(lines)
    if headers is True:
        df.columns = columns
    return df


def read_mother_ids(df):
    mother_ids = df.loc[:, "ID"].drop_duplicates().reset_index(drop=True)
    return mother_ids


def get_children(df, mother_id):
    children = df.loc[df['ID'] == mother_id].reset_index(drop=True)
    return children


def get_pages(csv: Union[str, pd.DataFrame], image_path: Optional[str] = None, split_by_mother: bool = False,
              multipred_classes: Optional[Tuple[List, List]] = None) -> List[Union[Page, List[Page]]]:
    if isinstance(csv, str):
        df = read_csv(csv, ";")
    elif isinstance(csv, pd.DataFrame):
        df = csv
    else:
        raise ValueError

    if image_path is None:
        image_path = "%s"
        check_path = False
    else:
        check_path = True

    if split_by_mother:
        pages = [[]]
        current_id = df.loc[0, u"Parent_ID"]
        for index, row in df.iterrows():
            path = image_path % row[u"Path"]

            if check_path and not os.path.isfile(path):
                UserWarning(f"Path does not exist: {path}")
                continue  # skip pages which have no image

            if multipred_classes is not None:
                multipred = generate_multipred(multipred_classes, row)
            else:
                multipred = None

            p = Page(index, row[u"Parent_ID"], path, row[u"Predicted"], row[u"Actual"], row[u"Page"], row[u"Text"],
                     multipred)
            if current_id == row[u"Parent_ID"]:
                pages[-1].append(p)
            else:
                current_id = row[u"Parent_ID"]
                pages.append([p])
    else:
        pages = []
        for index, row in df.iterrows():
            path = image_path % row[u"Path"]

            if check_path and not os.path.isfile(path):
                UserWarning(f"Path does not exist: {path}")
                continue

            if multipred_classes is not None:
                multipred = generate_multipred(multipred_classes, row)
            else:
                multipred = None

            p = Page(index, row[u"Parent_ID"], path, row[u"Predicted"], row[u"Actual"], row[u"Page"], row[u"Text"],
                     multipred)
            pages.append(p)
    return pages


def iter_dict(d: dict, lst=None):
    if lst is None:
        lst = []
    for key in d:
        val = d[key]
        if isinstance(val, dict):
            iter_dict(val, lst)
        elif isinstance(val, Number):
            lst.append(val)
    return lst


def generate_multipred(multipred_classes, row):
    image_classes = multipred_classes[0]
    text_classes = multipred_classes[1]
    st = row["Multipred"].replace("\"\"\"", "")
    # print(st)
    wow = {}
    if st[0] == "{":
        js = json.loads(st)
        # print(js)

        wow['ImagePredictions'] = {ic: 0.0 for ic in image_classes}
        if 'ImagePredictions' in js:
            for i in range(len(js['ImagePredictions'])):
                wow['ImagePredictions'][js['ImagePredictions'][i]['TagName']] = js['ImagePredictions'][i][
                    'Probability']

        # wow['skjema'] = 0.0
        # if 'SchemaPrediction' in js:
        #     wow[js['SchemaPrediction']['TagName']] = js['SchemaPrediction']['Probability']

        wow['TextPredictions'] = {tc: 0.0 for tc in text_classes}
        if 'TextPredictions' in js:
            for i in range(len(js['TextPredictions'])):
                if js['TextPredictions'][i]['TagName'] not in text_classes:
                    raise ValueError(f"{js['TextPredictions'][i]['TagName']} is not in text class list")
                wow['TextPredictions'][js['TextPredictions'][i]['TagName']] = js['TextPredictions'][i][
                    'Probability']

        return iter_dict(wow)
    else:
        return None


def txt_to_pages(path, images):
    f = open(path)

    txt = f.read()

    split = [s.split() for s in txt.split("\n")]

    pages = []
    i = 0
    for row in split:
        if len(row) == 2:
            pages.append(Page(i, 0, images % row[0], None, row[1], i, None))
            i += 1

    return pages


def classify_pages_data(pages: List[Page]):
    paths = [p.path for p in pages]
    labels = [p.true_class for p in pages]

    return paths, labels


def get_dataset_classify(pages, classes, batch, h, w, c, epochs=None, shuffle=None, buffer_size=BUFFER_SIZE,
                         parallel_calls=PARALLEL_CALLS):
    paths, labels = classify_pages_data(pages)

    all_path_labels = [classes[l.replace('"""', "").strip().lower()] for l in labels]

    paths = tf.data.Dataset.from_tensor_slices(paths)
    labels = tf.data.Dataset.from_tensor_slices(tf.cast(all_path_labels, tf.int64))
    data = tf.data.Dataset.zip((paths, labels))

    def load_and_preprocess_from_path_label(path, label):
        return process_image(path, h, w, c), one_hot(label, len(classes))

    data = data.map(load_and_preprocess_from_path_label, num_parallel_calls=parallel_calls)
    if shuffle is not None:
        data = data.shuffle(buffer_size=shuffle)
    data = data.batch(batch)
    data = data.prefetch(buffer_size=buffer_size)
    if epochs is not None:
        data = data.repeat(epochs)

    return data
