from nltk.stem import SnowballStemmer
from tensorflow.python.keras.preprocessing.text import text_to_word_sequence


def tokenize(text):
    """
    Divides the input string into words, splitting on any non-alphabethic
    characters and removes words shorter than three characters.
    :param text: The string to be divided.
    :return: An array of the remaining words.
    """
    return remove_short(
        text_to_word_sequence(text, filters='!' + '"' + "'" + '\n' + '#$%—&()*+,-./:;<=>?@[\]^_`{|}~1234567890'))


def remove_short(words):
    temp = []
    for word in words:
        if len(word) > 2:
            temp.append(word)
    return temp


def stem(words, language=None):
    """
    Finds the root of the words in an array in the desired language.
    :param language: The language of the words.
    :param words: An array with the words to stem.
    :return: An array with the roots of the input.
    """
    if language == "en":
        ss = SnowballStemmer("english")
    else:
        ss = SnowballStemmer("norwegian")
    res = []
    for w in words:
        res.append(ss.stem(w))
    return res


def remove_stop(text, stopwords=None, remove_short_words=False, num=0):
    """
    Removes short words and stopwords from the string.
    :param text: An array of words to check.
    :param remove_short_words: An optional boolean value determining
            whether to remove short words. Defaults to False
    :param num: The minimum number of letters in the words. Defaults to zero.
    :return: The trimmed string.
    """

    if stopwords is None:
        stopwords = []

    if remove_short_words:
        res = []
        for word in text:
            if len(word) > num:
                res.append(word)
        text = res

    res = [word for word in text if word not in stopwords]

    return res


def preprocess(text, tokenization=True, stop_word=True, stemming=True):
    """
    Combines multiple pre-processing algorithms.
    1) Language detection.
    2) Tokenization to split the text into words and remove non-alphabetic characters.
    3) Removing stop-words for the language that was detected in step 1.
    4) Stem the remaining word (Remove the ending of words, so only the root is left.
    :param stemming: An optional boolean value determining whether to stem the string. Defaults to true.
    :param stop_word: An optional boolean value determining whether to remove stop-words. Defaults to true.
    :param tokenization: An optional boolean value determining whether to tokenize the string. Defaults to true.
    :param text: The text string to be processed.
    :return: The processed text.
    """
    # # Language detection
    # try:
    #     lang = detect_lang(text)
    # except:
    #     lang = "no"

    # Tokenization
    if tokenization:
        text = tokenize(text)

    # Stop-word removal
    if stop_word:
        filename = "../res/stopWords_no"
        with open(filename, encoding='utf-8') as file:
            s = file.read()
            stopwords = s.split("\n")

        text = remove_stop(text, stopwords, True, 2)

    # Stemming
    if stemming:
        text = stem(text, 'no')

    return " ".join(text)
